/*
 *  linux/arch/arm/mm/extable.c
 */

#include <linux/config.h>
#include <linux/module.h>
#include <asm/uaccess.h>

const struct exception_table_entry *
search_extable(const struct exception_table_entry *first,
               const struct exception_table_entry *last,
               unsigned long value)
{
        while (first <= last) {
                const struct exception_table_entry *mid;
                long diff;

                mid = (last - first) / 2 + first;
                diff = mid->insn - value;
                if (diff == 0)
                        return mid;
                else if (diff < 0)
                        first = mid+1;
                else
                        last = mid-1;
        }
        return NULL;
}

int fixup_exception(struct pt_regs *regs)
{
        const struct exception_table_entry *fixup;

        fixup = search_exception_tables(instruction_pointer(regs));
        if (fixup)
                regs->ARM_pc = fixup->fixup | PSR_I_BIT | MODE_SVC26;

        return fixup != NULL;
}

