ID("CSC0000", "Crystal Semiconductor CS423x sound -- SB/WSS/OPL3 emulation")
ID("CSC0010", "Crystal Semiconductor CS423x sound -- control")
ID("CSC0001", "Crystal Semiconductor CS423x sound -- joystick")
ID("CSC0003", "Crystal Semiconductor CS423x sound -- MPU401")
ID("IBM3780", "IBM pointing device")
ID("IBM0071", "IBM infrared communications device")
ID("IBM3760", "IBM DSP")
ID("NSC6001", "National Semiconductor Serial Port with Fast IR")
ID("PNP0000", "AT Interrupt Controller")
ID("PNP0001", "EISA Interrupt Controller")
ID("PNP0002", "MCA Interrupt Controller")
ID("PNP0003", "APIC")
ID("PNP0004", "Cyrix SLiC MP Interrupt Controller")
ID("PNP0100", "AT Timer")
ID("PNP0101", "EISA Timer")
ID("PNP0102", "MCA Timer")
ID("PNP0200", "AT DMA Controller")
ID("PNP0201", "EISA DMA Controller")
ID("PNP0202", "MCA DMA Controller")
ID("PNP0300", "IBM PC/XT keyboard controller (83-key)")
ID("PNP0301", "IBM PC/AT keyboard controller (86-key)")
ID("PNP0302", "IBM PC/XT keyboard controller (84-key)")
ID("PNP0303", "IBM Enhanced (101/102-key, PS/2 mouse support)")
ID("PNP0304", "Olivetti Keyboard (83-key)")
ID("PNP0305", "Olivetti Keyboard (102-key)")
ID("PNP0306", "Olivetti Keyboard (86-key)")
ID("PNP0307", "Microsoft Windows(R) Keyboard")
ID("PNP0308", "General Input Device Emulation Interface (GIDEI) legacy")
ID("PNP0309", "Olivetti Keyboard (A101/102 key)")
ID("PNP030a", "AT&T 302 Keyboard")
ID("PNP0320", "Japanese 106-key keyboard A01")
ID("PNP0321", "Japanese 101-key keyboard")
ID("PNP0322", "Japanese AX keyboard")
ID("PNP0323", "Japanese 106-key keyboard 002/003")
ID("PNP0324", "Japanese 106-key keyboard 001")
ID("PNP0325", "Japanese Toshiba Desktop keyboard")
ID("PNP0326", "Japanese Toshiba Laptop keyboard")
ID("PNP0327", "Japanese Toshiba Notebook keyboard")
ID("PNP0340", "Korean 84-key keyboard")
ID("PNP0341", "Korean 86-key keyboard")
ID("PNP0342", "Korean Enhanced keyboard")
ID("PNP0343", "Korean Enhanced keyboard 101b")
ID("PNP0343", "Korean Enhanced keyboard 101c")
ID("PNP0344", "Korean Enhanced keyboard 103")
ID("PNP0400", "Standard LPT printer port")
ID("PNP0401", "ECP printer port")
ID("PNP0500", "Standard PC COM port")
ID("PNP0501", "16550A-compatible COM port")
ID("PNP0502", "Multiport serial device (non-intelligent 16550)")
ID("PNP0510", "Generic IRDA-compatible device")
ID("PNP0511", "Generic IRDA-compatible device")
ID("PNP0600", "Generic ESDI/IDE/ATA compatible hard disk controller")
ID("PNP0601", "Plus Hardcard II")
ID("PNP0602", "Plus Hardcard IIXL/EZ")
ID("PNP0603", "Generic IDE supporting Microsoft Device Bay Specification")
ID("PNP0700", "PC standard floppy disk controller")
ID("PNP0701", "Standard floppy controller supporting MS Device Bay Spec")
ID("PNP0802", "Microsoft Sound System or Compatible Device (obsolete)")
ID("PNP0900", "VGA Compatible")
ID("PNP0901", "Video Seven VRAM/VRAM II/1024i")
ID("PNP0902", "8514/A Compatible")
ID("PNP0903", "Trident VGA")
ID("PNP0904", "Cirrus Logic Laptop VGA")
ID("PNP0905", "Cirrus Logic VGA")
ID("PNP0906", "Tseng ET4000")
ID("PNP0907", "Western Digital VGA")
ID("PNP0908", "Western Digital Laptop VGA")
ID("PNP0909", "S3 Inc. 911/924")
ID("PNP090a", "ATI Ultra Pro/Plus (Mach 32)")
ID("PNP090b", "ATI Ultra (Mach 8)")
ID("PNP090c", "XGA Compatible")
ID("PNP090d", "ATI VGA Wonder")
ID("PNP090e", "Weitek P9000 Graphics Adapter")
ID("PNP090f", "Oak Technology VGA")
ID("PNP0910", "Compaq QVision")
ID("PNP0911", "XGA/2")
ID("PNP0912", "Tseng Labs W32/W32i/W32p")
ID("PNP0913", "S3 Inc. 801/928/964")
ID("PNP0914", "Cirrus Logic 5429/5434 (memory mapped)")
ID("PNP0915", "Compaq Advanced VGA (AVGA)")
ID("PNP0916", "ATI Ultra Pro Turbo (Mach64)")
ID("PNP0917", "Reserved by Microsoft")
ID("PNP0918", "Matrox MGA")
ID("PNP0919", "Compaq QVision 2000")
ID("PNP091a", "Tseng W128")
ID("PNP0930", "Chips & Technologies Super VGA")
ID("PNP0931", "Chips & Technologies Accelerator")
ID("PNP0940", "NCR 77c22e Super VGA")
ID("PNP0941", "NCR 77c32blt")
ID("PNP09ff", "Plug and Play Monitors (VESA DDC)")
ID("PNP0a00", "ISA Bus")
ID("PNP0a01", "EISA Bus")
ID("PNP0a02", "MCA Bus")
ID("PNP0a03", "PCI Bus")
ID("PNP0a04", "VESA/VL Bus")
ID("PNP0a05", "Generic ACPI Bus")
ID("PNP0a06", "Generic ACPI Extended-IO Bus (EIO bus)")
ID("PNP0800", "AT-style speaker sound")
ID("PNP0b00", "AT Real-Time Clock")
ID("PNP0c00", "Plug and Play BIOS (only created by the root enumerator)")
ID("PNP0c01", "System Board")
ID("PNP0c02", "Reserved Motherboard Resources")
ID("PNP0c03", "Plug and Play BIOS Event Notification Interrupt")
ID("PNP0c04", "Math Coprocessor")
ID("PNP0c05", "APM BIOS (Version independent)")
ID("PNP0c06", "Reserved for identification of early Plug and Play BIOS implementation.")
ID("PNP0c07", "Reserved for identification of early Plug and Play BIOS implementation.")
ID("PNP0c08", "ACPI system board hardware")
ID("PNP0c09", "ACPI Embedded Controller")
ID("PNP0c0a", "ACPI Control Method Battery")
ID("PNP0c0b", "ACPI Fan")
ID("PNP0c0c", "ACPI power button device")
ID("PNP0c0d", "ACPI lid device")
ID("PNP0c0e", "ACPI sleep button device")
ID("PNP0c0f", "PCI interrupt link device")
ID("PNP0c10", "ACPI system indicator device")
ID("PNP0c11", "ACPI thermal zone")
ID("PNP0c12", "Device Bay Controller")
ID("PNP0c13", "Plug and Play BIOS (used when ACPI mode cannot be used)")
ID("PNP0e00", "Intel 82365-Compatible PCMCIA Controller")
ID("PNP0e01", "Cirrus Logic CL-PD6720 PCMCIA Controller")
ID("PNP0e02", "VLSI VL82C146 PCMCIA Controller")
ID("PNP0e03", "Intel 82365-compatible CardBus controller")
ID("PNP0f00", "Microsoft Bus Mouse")
ID("PNP0f01", "Microsoft Serial Mouse")
ID("PNP0f02", "Microsoft InPort Mouse")
ID("PNP0f03", "Microsoft PS/2-style Mouse")
ID("PNP0f04", "Mouse Systems Mouse")
ID("PNP0f05", "Mouse Systems 3-Button Mouse (COM2)")
ID("PNP0f06", "Genius Mouse (COM1)")
ID("PNP0f07", "Genius Mouse (COM2)")
ID("PNP0f08", "Logitech Serial Mouse")
ID("PNP0f09", "Microsoft BallPoint Serial Mouse")
ID("PNP0f0a", "Microsoft Plug and Play Mouse")
ID("PNP0f0b", "Microsoft Plug and Play BallPoint Mouse")
ID("PNP0f0c", "Microsoft-compatible Serial Mouse")
ID("PNP0f0d", "Microsoft-compatible InPort-compatible Mouse")
ID("PNP0f0e", "Microsoft-compatible PS/2-style Mouse")
ID("PNP0f0f", "Microsoft-compatible Serial BallPoint-compatible Mouse")
ID("PNP0f10", "Texas Instruments QuickPort Mouse")
ID("PNP0f11", "Microsoft-compatible Bus Mouse")
ID("PNP0f12", "Logitech PS/2-style Mouse")
ID("PNP0f13", "PS/2 Port for PS/2-style Mice")
ID("PNP0f14", "Microsoft Kids Mouse")
ID("PNP0f15", "Logitech bus mouse")
ID("PNP0f16", "Logitech SWIFT device")
ID("PNP0f17", "Logitech-compatible serial mouse")
ID("PNP0f18", "Logitech-compatible bus mouse")
ID("PNP0f19", "Logitech-compatible PS/2-style Mouse")
ID("PNP0f1a", "Logitech-compatible SWIFT Device")
ID("PNP0f1b", "HP Omnibook Mouse")
ID("PNP0f1c", "Compaq LTE Trackball PS/2-style Mouse")
ID("PNP0f1d", "Compaq LTE Trackball Serial Mouse")
ID("PNP0f1e", "Microsoft Kids Trackball Mouse")
ID("PNP8001", "Novell/Anthem NE3200")
ID("PNP8004", "Compaq NE3200")
ID("PNP8006", "Intel EtherExpress/32")
ID("PNP8008", "HP EtherTwist EISA LAN Adapter/32 (HP27248A)")
ID("PNP8065", "Ungermann-Bass NIUps or NIUps/EOTP")
ID("PNP8072", "DEC (DE211) EtherWorks MC/TP")
ID("PNP8073", "DEC (DE212) EtherWorks MC/TP_BNC")
ID("PNP8078", "DCA 10 Mb MCA")
ID("PNP8074", "HP MC LAN Adapter/16 TP (PC27246)")
ID("PNP80c9", "IBM Token Ring")
ID("PNP80ca", "IBM Token Ring II")
ID("PNP80cb", "IBM Token Ring II/Short")
ID("PNP80cc", "IBM Token Ring 4/16Mbs")
ID("PNP80d3", "Novell/Anthem NE1000")
ID("PNP80d4", "Novell/Anthem NE2000")
ID("PNP80d5", "NE1000 Compatible")
ID("PNP80d6", "NE2000 Compatible")
ID("PNP80d7", "Novell/Anthem NE1500T")
ID("PNP80d8", "Novell/Anthem NE2100")
ID("PNP80dd", "SMC ARCNETPC")
ID("PNP80de", "SMC ARCNET PC100, PC200")
ID("PNP80df", "SMC ARCNET PC110, PC210, PC250")
ID("PNP80e0", "SMC ARCNET PC130/E")
ID("PNP80e1", "SMC ARCNET PC120, PC220, PC260")
ID("PNP80e2", "SMC ARCNET PC270/E")
ID("PNP80e5", "SMC ARCNET PC600W, PC650W")
ID("PNP80e7", "DEC DEPCA")
ID("PNP80e8", "DEC (DE100) EtherWorks LC")
ID("PNP80e9", "DEC (DE200) EtherWorks Turbo")
ID("PNP80ea", "DEC (DE101) EtherWorks LC/TP")
ID("PNP80eb", "DEC (DE201) EtherWorks Turbo/TP")
ID("PNP80ec", "DEC (DE202) EtherWorks Turbo/TP_BNC")
ID("PNP80ed", "DEC (DE102) EtherWorks LC/TP_BNC")
ID("PNP80ee", "DEC EE101 (Built-In)")
ID("PNP80ef", "DECpc 433 WS (Built-In)")
ID("PNP80f1", "3Com EtherLink Plus")
ID("PNP80f3", "3Com EtherLink II or IITP (8 or 16-bit)")
ID("PNP80f4", "3Com TokenLink")
ID("PNP80f6", "3Com EtherLink 16")
ID("PNP80f7", "3Com EtherLink III")
ID("PNP80f8", "3Com Generic Etherlink Plug and Play Device")
ID("PNP80fb", "Thomas Conrad TC6045")
ID("PNP80fc", "Thomas Conrad TC6042")
ID("PNP80fd", "Thomas Conrad TC6142")
ID("PNP80fe", "Thomas Conrad TC6145")
ID("PNP80ff", "Thomas Conrad TC6242")
ID("PNP8100", "Thomas Conrad TC6245")
ID("PNP8105", "DCA 10 MB")
ID("PNP8106", "DCA 10 MB Fiber Optic")
ID("PNP8107", "DCA 10 MB Twisted Pair")
ID("PNP8113", "Racal NI6510")
ID("PNP811c", "Ungermann-Bass NIUpc")
ID("PNP8120", "Ungermann-Bass NIUpc/EOTP")
ID("PNP8123", "SMC StarCard PLUS (WD/8003S)")
ID("PNP8124", "SMC StarCard PLUS With On Board Hub (WD/8003SH)")
ID("PNP8125", "SMC EtherCard PLUS (WD/8003E)")
ID("PNP8126", "SMC EtherCard PLUS With Boot ROM Socket (WD/8003EBT)")
ID("PNP8127", "SMC EtherCard PLUS With Boot ROM Socket (WD/8003EB)")
ID("PNP8128", "SMC EtherCard PLUS TP (WD/8003WT)")
ID("PNP812a", "SMC EtherCard PLUS 16 With Boot ROM Socket (WD/8013EBT)")
ID("PNP812d", "Intel EtherExpress 16 or 16TP")
ID("PNP812f", "Intel TokenExpress 16/4")
ID("PNP8130", "Intel TokenExpress MCA 16/4")
ID("PNP8132", "Intel EtherExpress 16 (MCA)")
ID("PNP8137", "Artisoft AE-1")
ID("PNP8138", "Artisoft AE-2 or AE-3")
ID("PNP8141", "Amplicard AC 210/XT")
ID("PNP8142", "Amplicard AC 210/AT")
ID("PNP814b", "Everex SpeedLink /PC16 (EV2027)")
ID("PNP8155", "HP PC LAN Adapter/8 TP (HP27245)")
ID("PNP8156", "HP PC LAN Adapter/16 TP (HP27247A)")
ID("PNP8157", "HP PC LAN Adapter/8 TL (HP27250)")
ID("PNP8158", "HP PC LAN Adapter/16 TP Plus (HP27247B)")
ID("PNP8159", "HP PC LAN Adapter/16 TL Plus (HP27252)")
ID("PNP815f", "National Semiconductor Ethernode *16AT")
ID("PNP8160", "National Semiconductor AT/LANTIC EtherNODE 16-AT3")
ID("PNP816a", "NCR Token-Ring 4 Mbs ISA")
ID("PNP816d", "NCR Token-Ring 16/4 Mbs ISA")
ID("PNP8191", "Olicom 16/4 Token-Ring Adapter")
ID("PNP81c3", "SMC EtherCard PLUS Elite (WD/8003EP)")
ID("PNP81c4", "SMC EtherCard PLUS 10T (WD/8003W)")
ID("PNP81c5", "SMC EtherCard PLUS Elite 16 (WD/8013EP)")
ID("PNP81c6", "SMC EtherCard PLUS Elite 16T (WD/8013W)")
ID("PNP81c7", "SMC EtherCard PLUS Elite 16 Combo (WD/8013EW or 8013EWC)")
ID("PNP81c8", "SMC EtherElite Ultra 16")
ID("PNP81e4", "Pure Data PDI9025-32 (Token Ring)")
ID("PNP81e6", "Pure Data PDI508+ (ArcNet)")
ID("PNP81e7", "Pure Data PDI516+ (ArcNet)")
ID("PNP81eb", "Proteon Token Ring (P1390)")
ID("PNP81ec", "Proteon Token Ring (P1392)")
ID("PNP81ed", "Proteon ISA Token Ring (1340)")
ID("PNP81ee", "Proteon ISA Token Ring (1342)")
ID("PNP81ef", "Proteon ISA Token Ring (1346)")
ID("PNP81f0", "Proteon ISA Token Ring (1347)")
ID("PNP81ff", "Cabletron E2000 Series DNI")
ID("PNP8200", "Cabletron E2100 Series DNI")
ID("PNP8209", "Zenith Data Systems Z-Note")
ID("PNP820a", "Zenith Data Systems NE2000-Compatible")
ID("PNP8213", "Xircom Pocket Ethernet II")
ID("PNP8214", "Xircom Pocket Ethernet I")
ID("PNP821d", "RadiSys EXM-10")
ID("PNP8227", "SMC 3000 Series")
ID("PNP8228", "SMC 91C2 controller")
ID("PNP8231", "Advanced Micro Devices AM2100/AM1500T")
ID("PNP8263", "Tulip NCC-16")
ID("PNP8277", "Exos 105")
ID("PNP828a", "Intel '595 based Ethernet")
ID("PNP828b", "TI2000-style Token Ring")
ID("PNP828c", "AMD PCNet Family cards")
ID("PNP828d", "AMD PCNet32 (VL version)")
ID("PNP8294", "IrDA Infrared NDIS driver (Microsoft-supplied)")
ID("PNP82bd", "IBM PCMCIA-NIC")
ID("PNP82c2", "Xircom CE10")
ID("PNP82c3", "Xircom CEM2")
ID("PNP8321", "DEC Ethernet (All Types)")
ID("PNP8323", "SMC EtherCard (All Types except 8013/A)")
ID("PNP8324", "ARCNET Compatible")
ID("PNP8326", "Thomas Conrad (All Arcnet Types)")
ID("PNP8327", "IBM Token Ring (All Types)")
ID("PNP8385", "Remote Network Access Driver")
ID("PNP8387", "RNA Point-to-point Protocol Driver")
ID("PNP8388", "Reserved for Microsoft Networking components")
ID("PNP8389", "Peer IrLAN infrared driver (Microsoft-supplied)")
ID("PNP8390", "Generic network adapter")
ID("PNPa002", "Future Domain 16-700 compatible controller")
ID("PNPa003", "Panasonic proprietary CD-ROM adapter (SBPro/SB16)")
ID("PNPa01b", "Trantor 128 SCSI Controller")
ID("PNPa01d", "Trantor T160 SCSI Controller")
ID("PNPa01e", "Trantor T338 Parallel SCSI controller")
ID("PNPa01f", "Trantor T348 Parallel SCSI controller")
ID("PNPa020", "Trantor Media Vision SCSI controller")
ID("PNPa022", "Always IN-2000 SCSI controller")
ID("PNPa02b", "Sony proprietary CD-ROM controller")
ID("PNPa02d", "Trantor T13b 8-bit SCSI controller")
ID("PNPa02f", "Trantor T358 Parallel SCSI controller")
ID("PNPa030", "Mitsumi LU-005 Single Speed CD-ROM controller + drive")
ID("PNPa031", "Mitsumi FX-001 Single Speed CD-ROM controller + drive")
ID("PNPa032", "Mitsumi FX-001 Double Speed CD-ROM controller + drive")
ID("PNPb000", "Sound Blaster 1.5 sound device")
ID("PNPb001", "Sound Blaster 2.0 sound device")
ID("PNPb002", "Sound Blaster Pro sound device")
ID("PNPb003", "Sound Blaster 16 sound device")
ID("PNPb004", "Thunderboard-compatible sound device")
ID("PNPb005", "Adlib-compatible FM synthesizer device")
ID("PNPb006", "MPU401 compatible")
ID("PNPb007", "Microsoft Windows Sound System-compatible sound device")
ID("PNPb008", "Compaq Business Audio")
ID("PNPb009", "Plug and Play Microsoft Windows Sound System Device")
ID("PNPb00a", "MediaVision Pro Audio Spectrum (Trantor SCSI enabled, Thunder Chip Disabled)")
ID("PNPb00b", "MediaVision Pro Audio 3D")
ID("PNPb00c", "MusicQuest MQX-32M")
ID("PNPb00d", "MediaVision Pro Audio Spectrum Basic (No Trantor SCSI, Thunder Chip Enabled)")
ID("PNPb00e", "MediaVision Pro Audio Spectrum (Trantor SCSI enabled, Thunder Chip Enabled)")
ID("PNPb00f", "MediaVision Jazz-16 chipset (OEM Versions)")
ID("PNPb010", "Auravision VxP500 chipset - Orchid Videola")
ID("PNPb018", "MediaVision Pro Audio Spectrum 8-bit")
ID("PNPb019", "MediaVision Pro Audio Spectrum Basic (no Trantor SCSI, Thunder chip Disabled)")
ID("PNPb020", "Yamaha OPL3-compatible FM synthesizer device")
ID("PNPb02f", "Joystick/Game port")
ID("PNPb000", "Compaq 14400 Modem (TBD)")
ID("PNPc001", "Compaq 2400/9600 Modem (TBD)")
ID("PNP0XXX", "Unknown System Device")
ID("PNP8XXX", "Unknown Network Adapter")
ID("PNPaXXX", "Unknown SCSI, Proprietary CD Adapter")
ID("PNPbXXX", "Unknown Multimedia Device")
ID("PNPcXXX", "Unknown Modem")

#undef ID
