/*
 * arch/ppc/platforms/menf1.h
 * 
 * Definitions for MEN F1 board support
 *
 * Author: Matt Porter <mporter@mvista.com>
 *
 * 2001 (c) MontaVista, Software, Inc.  This file is licensed under
 * the terms of the GNU General Public License version 2.  This program
 * is licensed "as is" without any warranty of any kind, whether express
 * or implied.
 */

#ifndef __PPC_PLATFORMS_MENF1_H
#define __PPC_PLATFORMS_MENF1_H

#define MENF1_NVRAM_AS0			0x70
#define MENF1_NVRAM_AS1			0x72
#define MENF1_NVRAM_DATA		0x71    

#define	MENF1_IDE0_BASE_ADDR		0x1f0
#define MENF1_IDE1_BASE_ADDR		0x170

#endif /* __PPC_PLATFORMS_MENF1_H */
