/*
 * include/linux/topology.h
 *
 * Written by: Matthew Dobson, IBM Corporation
 *
 * Copyright (C) 2002, IBM Corp.
 *
 * All rights reserved.          
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, GOOD TITLE or
 * NON INFRINGEMENT.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Send feedback to <colpatch@us.ibm.com>
 */
#ifndef _LINUX_TOPOLOGY_H
#define _LINUX_TOPOLOGY_H

#include <linux/bitops.h>
#include <linux/mmzone.h>
#include <linux/smp.h>

#include <asm/topology.h>

#ifndef nr_cpus_node
#define nr_cpus_node(node)	(hweight_long(node_to_cpumask(node)))
#endif

static inline int __next_node_with_cpus(int node)
{
	do
		++node;
	while (node < numnodes && !nr_cpus_node(node));
	return node;
}

#define for_each_node_with_cpus(node) \
	for (node = 0; node < numnodes; node = __next_node_with_cpus(node))

#endif /* _LINUX_TOPOLOGY_H */
